from jaypore_ci.config import Version

V = Version.parse
NEW = "🎁"
CHANGE = "⚙️"
BUGFIX = "🐞"

version_map = {
    V("0.2.30"): {
        "changes": [
            (
                f"{NEW}: You can pass arbitrary arguments to the `docker run` "
                "command simply by using the `executor_kwargs` argument while "
                "defining the job. Read more in `Passing extra_hosts and other "
                "arguments to docker`_."
            ),
            f"{NEW}: SSH remotes are now compatible with Jaypore CI.",
        ],
        "instructions": [],
    },
    V("0.2.29"): {
        "changes": [
            (
                f"{BUGFIX}: When gitea token does not have enough scope log"
                " correctly and exit"
            )
        ],
        "instructions": [],
    },
    V("0.2.28"): {
        "changes": [
            (
                f"{BUGFIX}: When there are multiple (push) remotes, Jaypore CI"
                " will pick the first one and use that."
            )
        ],
        "instructions": [],
    },
    V("0.2.27"): {
        "changes": [
            (
                f"{NEW}: Jobs older than 1 week will be removed before starting"
                " a new pipeline."
            )
        ],
        "instructions": [],
    },
    V("0.2.26"): {
        "changes": [
            (
                f"{CHANGE}: The Dockerfile inside `cicd/Dockerfile` now"
                " requires a build arg that specifies the version of Jaypore CI"
                " to install."
            ),
        ],
        "instructions": [
            "Please run the Jaypore CI setup once again.",
        ],
    },
    V("0.2.25"): {
        "changes": [
            (
                f"{NEW}: A dockerfile is now used to send context of the"
                " codebase to the docker daemon instead of directly mounting the"
                " code. This allows us to easily use remote systems for jobs"
            )
        ],
        "instructions": [],
    },
}
assert all(
    line.startswith(NEW) or line.startswith(CHANGE) or line.startswith(BUGFIX)
    for log in version_map.values()
    for line in log["changes"]
), "All change lines must start with one of NEW/CHANGE/BUGFIX"



import os
import tomllib
from typing import NamedTuple
import importlib.metadata
from pathlib import Path


class Version(NamedTuple):
    major: int
    minor: int
    patch: int
    trail: str = None

    def __repr__(self):
        if self.trail:
            return f"{self.major}.{self.minor}.{self.patch}-{self.trail}"
        return f"{self.major}.{self.minor}.{self.patch}"

    def __str__(self):
        return self.__repr__()

    @classmethod
    def parse(cls, inp: str) -> "Version":
        if inp is None or inp == "":
            return None
        trail = None
        major, minor, patch = inp.split(".")
        major = major[1:] if major[0].lower() == "v" else major
        assert major.isdigit()
        assert minor.isdigit()
        if "-" in patch:
            patch, trail = patch.split("-", 1)
            assert patch.isdigit()
        return cls(major=int(major), minor=int(minor), patch=int(patch), trail=trail)


def get_version() -> Version:
    try:
        return Version.parse(importlib.metadata.version(__package__ or __name__))
    except importlib.metadata.PackageNotFoundError:
        try:
            with open(
                (Path(__file__) / "../../pyproject.toml").resolve(),
                "rb",
            ) as fl:
                data = tomllib.load(fl)
            return Version.parse(data["tool"]["poetry"]["version"])
        except FileNotFoundError:
            return None


class Const(NamedTuple):
    expected_version: Version = Version.parse(
        os.environ.get("EXPECTED_JAYPORECI_VERSION")
    )
    version: Version = get_version()


const = Const()



"""
Defines interfaces for remotes and executors.

Currently only gitea and docker are supported as remote and executor
respectively.
"""
from enum import Enum
from pathlib import Path
from urllib.parse import urlparse
from typing import NamedTuple, List


class TriggerFailed(Exception):
    "Failure to trigger a job"


class RemoteApiFailed(Exception):
    "Failure while working with a remote"


class JobStatus(NamedTuple):
    is_running: bool
    exit_code: int
    logs: str
    started_at: str
    finished_at: str


class Status(Enum):
    "Each pipeline can ONLY be in any one of these statuses"
    PENDING = 10
    RUNNING = 30
    FAILED = 40
    PASSED = 50
    TIMEOUT = 60
    SKIPPED = 70


class RemoteInfo(NamedTuple):
    """
    Holds information about the remote irrespective of if the remote was ssh or
    https.
    """

    netloc: str
    owner: str
    repo: str
    original: str

    @classmethod
    def parse(cls, remote: str) -> "RemoteInfo":
        """
        Given a git remote url string, parses and breaks down information
        contained in the url.

        Works with the following formats:

            ssh://git@gitea.arjoonn.com:arjoonn/jaypore_ci.git
            ssh+git://git@gitea.arjoonn.com:arjoonn/jaypore_ci.git

            git@gitea.arjoonn.com:arjoonn/jaypore_ci.git
            git@gitea.arjoonn.com:arjoonn/jaypore_ci.git

            https://gitea.arjoonn.com/midpath/jaypore_ci.git
            http://gitea.arjoonn.com/midpath/jaypore_ci.git
        """
        original = remote
        if (
            ("ssh://" in remote or "ssh+git://" in remote or "://" not in remote)
            and "@" in remote
            and remote.endswith(".git")
        ):
            _, remote = remote.split("@")
            netloc, path = remote.split(":")
            owner, repo = path.split("/")
            return RemoteInfo(
                netloc=netloc,
                owner=owner,
                repo=repo.replace(".git", ""),
                original=original,
            )
        url = urlparse(remote)
        return RemoteInfo(
            netloc=url.netloc,
            owner=Path(url.path).parts[1],
            repo=Path(url.path).parts[2].replace(".git", ""),
            original=original,
        )


class Repo:
    """
    Contains information about the current VCS repo.
    """

    def __init__(self, sha: str, branch: str, remote: str, commit_message: str):
        self.sha: str = sha
        self.branch: str = branch
        self.remote: str = remote
        self.commit_message: str = commit_message

    def files_changed(self, target: str) -> List[str]:
        """
        Returns list of file paths that have changed between current sha and
        target.
        """
        raise NotImplementedError()

    @classmethod
    def from_env(cls) -> "Repo":
        """
        Creates a :class:`~jaypore_ci.interfaces.Repo` instance
        from the environment and git repo on disk.
        """
        raise NotImplementedError()


class Executor:
    """
    An executor is something used to run a job.
    It could be docker / podman / shell etc.
    """

    def run(self, job: "Job") -> str:
        "Run a job and return it's ID"
        raise NotImplementedError()

    def __init__(self):
        self.pipe_id = None
        self.pipeline = None

    def set_pipeline(self, pipeline: "Pipeline") -> None:
        """Set the current pipeline to the given one."""
        self.pipe_id = id(pipeline)
        self.pipeline = pipeline

    def setup(self) -> None:
        """
        This function is meant to perform any work that should be done before
        running any jobs.
        """

    def teardown(self) -> None:
        """
        On exit the executor must clean up any pending / stuck / zombie jobs that are still there.
        """

    def get_status(self, run_id: str) -> JobStatus:
        """
        Returns the status of a given run.
        """
        raise NotImplementedError()


class Remote:
    """
    Something that allows us to show other people the status of the CI job.
    It could be gitea / github / gitlab / email system.
    """

    def __init__(self, *, sha, branch):
        self.sha = sha
        self.branch = branch

    def publish(self, report: str, status: str):
        """
        Publish this report somewhere.
        """
        raise NotImplementedError()

    def setup(self) -> None:
        """
        This function is meant to perform any work that should be done before
        running any jobs.
        """

    def teardown(self) -> None:
        """
        This function will be called once the pipeline is finished.
        """

    @classmethod
    def from_env(cls, *, repo: "Repo"):
        """
        This function should create a Remote instance from the given environment.
        It can read git information / look at environment variables etc.
        """
        raise NotImplementedError()


class Reporter:
    """
    Something that generates the status of a pipeline.

    It can be used to generate reports in markdown, plaintext, html, pdf etc.
    """

    def render(self, pipeline: "Pipeline") -> str:
        """
        Render a report for the pipeline.
        """
        raise NotImplementedError()
